/* MODLATCH.C -- source for a tile importer/exporter for Keen 1, 2, and 3.
** Copyright (c)2002 Andrew Durdin. Greetings and thanks to Anders Gavare.
**
** This code is here for your reference only; if you want to reuse any of it
** make sure you get permission first! You can contact me at andy@durdin.net
**
** Compiled with DJGPP.
**
** Revision history:
** 25 March 2002: (implicit) Version 1.0
**		Initial release, DJGPP+Allegro
** 27 March 2002: Version 1.1
**		Added LZ routines and Keen 1 support, and BMP i/o routines so it
**		compiles under plain DJGPP.
**  1 April 2002:	Renamed to MODKEEN version 1.0
**		Full support for editing sprites and bitmaps in all three.!!!
*/

// I'm going to totally change the way this works: put in a robust option
// parsing scheme; load all the data into memory, then export/import from
// memory. So I'll store all the stuff in memory as BITMAP16s, and write
// them from there back to EGA*.

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <stdlib.h>

#include "lz.h"
#include "bmp16.h"


#pragma pack(1)
/* Define the structures for EGAHEAD */
typedef struct {
	long LatchPlaneSize;		//Size of one plane of latch data
   long SpritePlaneSize;	//Size of one plane of sprite data
   long OffBitmapTable;		//Offset in EGAHEAD to bitmap table
   long OffSpriteTable;		//Offset in EGAHEAD to sprite table
   short Num8Tiles;			//Number of 8x8 tiles
   long Off8Tiles;			//Offset of 8x8 tiles (relative to plane data)
   short Num32Tiles;			//Number of 32x32 tiles (always 0)
   long Off32Tiles;			//Offset of 32x32 tiles (relative to plane data)
   short Num16Tiles;			//Number of 16x16 tiles
   long Off16Tiles;			//Offset of 16x16 tiles (relative to plane data)
   short NumBitmaps;			//Number of bitmaps in table
   long OffBitmaps;			//Offset of bitmaps (relative to plane data)
   short NumSprites;			//Number of sprites
   long OffSprites;			//Offset of sprites (relative to plane data)
   short Compressed;			//(Keen 1 only) Nonzero: LZ compressed data
} EgaHead;

typedef struct {
	short Width;
   short Height;
   long Offset;
   char Name[8];
} BitmapHead;

typedef struct {
	short Width;
   short Height;
   short OffsetDelta;
   short OffsetParas;
   short Rx1, Ry1, Rx2, Ry2;
   char Name[16];
} SpriteHead;
#pragma pack()

void export( EgaHead *head, BitmapHead *bmphead, SpriteHead *sprhead,
		char *latchdata, char *spritedata, char *basepath, char *listfilename, int episode );
void import( EgaHead *head, BitmapHead *bmphead, SpriteHead *sprhead,
		BITMAP16 *fontbmp, BITMAP16 *tilebmp, BITMAP16 **bitmaps, BITMAP16 **sprites,
		FILE *headfile, FILE *latchfile, FILE *spritefile );

void encode_plane( int plane, char *pbits, BITMAP16 *bmp, int bx, int by, int w, int h );
void decode_plane( int cmask, FILE *fout, BITMAP16 *bmp, int bx, int by,
	int w, int h );

char *readline( char *linebuf, char *tag, FILE *f );
char *strtrim( char *buf );


int main( int argc, char *argv[] ) {
	char basepath[PATH_MAX], listfile[PATH_MAX];
   char filename[PATH_MAX], bakfilename[PATH_MAX];
   int episode;
   char command;
	EgaHead head;
   BitmapHead *bmphead;
   SpriteHead *sprhead;
	FILE *headfile = NULL;
	FILE *latchfile = NULL;
	FILE *spritefile = NULL;
   long latchlength, sprlength;
   char *latchdata, *spritedata;

   /* Check that we have enough parameters */
   if( argc < 4 ) {
		printf(  "Use: MODKEEN command episode listfile\n"
	   			"  command is either 'e' or 'i' to export or import.\n"
               "  episode is the episode number to work with.\n"
               "  listfile is the list file containing the info.\n"
               "  MODKEEN assumes that all the source bitmaps will\n"
               "  be in the same directory as listfile, and that\n"
               "  the EGA*.CK? files will be in the current directory.\n\n"
               "Examples:\n"
               "\tMODKEEN e 3 C:\\KEEN3\\k3data.lst\n"
               "\tMODKEEN i 2 C:\\KEEN2\\k2data.lst\n\n" );
      return 1;
	}

   /* Determine the episode we're working with */
   episode = atoi( argv[2] );
   if( episode != 1 && episode != 2 && episode != 3 ) {
      printf( "ERROR: MODKEEN will only work for Keen episodes 1, 2, and 3!\n" );
      return 1;
   }

   /* Find the base path from the list file name */
   strcpy( listfile, argv[3] );
   strcpy( basepath, listfile );
	if( strrchr( basepath, '\\' ) )
	   *(strrchr( basepath, '\\' ) + 1) = 0;
	else
		basepath[0] = 0;

	/* Now work out what we need to do */
	command = tolower( argv[1][0] );
   if( command == 'e' ) {
   	/* Export all data */

      /* Open EGAHEAD */
      sprintf( filename, "EGAHEAD.CK%c", episode + '0' );
      headfile = fopen( filename, "rb" );
      if( !headfile ) {
      	printf( "ERROR: Can't open %s!\n", filename );
         return 1;
      }

      /* Read all the header data */
      fread( &head, sizeof( EgaHead ), 1, headfile );

      /* Allocate space for bitmap and sprite tables */
      bmphead = (BitmapHead *)malloc( head.NumBitmaps * sizeof( BitmapHead ) );
   	sprhead = (SpriteHead *)malloc( head.NumSprites * 4 * sizeof( SpriteHead ) );
   	if( !bmphead || !sprhead ) {
      	free( bmphead );
         free( sprhead );
			fclose( headfile );
         printf( "ERROR: Not enough memory to create tables!\n" );
         return 1;
   	}

      /* Read the bitmap and sprite tables */
      fseek( headfile, head.OffBitmapTable, SEEK_SET );
      fread( bmphead, sizeof( BitmapHead ), head.NumBitmaps, headfile );
      fseek( headfile, head.OffSpriteTable, SEEK_SET );
      fread( sprhead, 4 * sizeof( SpriteHead ), head.NumSprites, headfile );

      /* Open the latch and sprite files and read them into memory */
      sprintf( filename, "EGALATCH.CK%c", episode + '0' );
		latchfile = fopen( filename, "rb" );
      if( !latchfile ) {
      	free( bmphead );
         free( sprhead );
      	fclose( headfile );
			printf( "ERROR: Cannot open %s!\n", filename );
         return 1;
		}
      sprintf( filename, "EGASPRIT.CK%c", episode + '0' );
		spritefile = fopen( filename, "rb" );
      if( !spritefile ) {
      	free( bmphead );
         free( sprhead );
      	fclose( headfile );
         fclose( latchfile );
			printf( "ERROR: Cannot open %s!\n", filename );
         return 1;
		}
		latchdata = malloc( head.LatchPlaneSize * 4 );
      spritedata = malloc( head.SpritePlaneSize * 5 );
      if( !latchdata || !spritedata ) {
      	free( latchdata ); free( spritedata );
      	free( bmphead ); free( sprhead );
      	fclose( headfile );
      	fclose( latchfile ); fclose( spritefile );
         printf( "ERROR: Not enough memory to load latch and sprite data!\n" );
         return 1;
      }

      if( episode == 1 && head.Compressed) {
      	fseek( latchfile, 6, SEEK_SET );
         fseek( spritefile, 6, SEEK_SET );
         lz_decompress( spritefile, spritedata );
         lz_decompress( latchfile, latchdata );
     	} else {
      	fread( latchdata, head.LatchPlaneSize * 4, 1, latchfile );
         fread( spritedata, head.SpritePlaneSize * 5, 1, spritefile );
		}

      /* Now export all the data to files and update the listfile */
		export( &head, bmphead, sprhead, latchdata, spritedata, basepath, listfile, episode );

      /* Clean up */
     	free( latchdata ); free( spritedata );
     	free( bmphead ); free( sprhead );
     	fclose( headfile );
     	fclose( latchfile ); fclose( spritefile );

      /* And success */
      return 0;
   } else if( command == 'i' ) {
   	/* Import all data */
		FILE *flist;
		int i, j;
		char linebuf[1000];
		char namebuf[20];
		BITMAP16 *fontbmp, *tilebmp;
		BITMAP16 **bitmaps;
		BITMAP16 **sprites;

		/* Open list file */
		flist = fopen( listfile, "ra" );
		if( !flist ) {
			printf( "Error opening listfile %s!\n", listfile );
			return 1;
		}

		/* Find out the episode */
		if( episode != atoi( readline( linebuf, "Episode", flist ) ) ) {
			printf( "ERROR: List file is for different episode!\n" );
			fclose( flist );
			return 1;
		}

		/* Read the font bitmap */
		readline( linebuf, "Font", flist );
		strtrim( linebuf );
		sprintf( filename, "%s%s", basepath, linebuf );
		fontbmp = bmp_load( filename );
		if( !fontbmp ) {
			fclose( flist );
			printf( "ERROR: Cannot load %s!\n", filename );
			return 1;
		}
		head.Num8Tiles = (fontbmp->w / 8) * (fontbmp->h / 8);

		/* Read the tiles bitmap */
		readline( linebuf, "Tiles", flist );
		strtrim( linebuf );
		sprintf( filename, "%s%s", basepath, linebuf );
		tilebmp = bmp_load( filename );
		if( !tilebmp ) {
			fclose( flist );
			bmp_free( fontbmp );
			printf( "ERROR: Cannot load %s!\n", filename );
			return 1;
		}
		head.Num16Tiles = (tilebmp->w / 16) * (tilebmp->h / 16);

		/* Read the bitmaps */
		head.NumBitmaps = atoi( readline( linebuf, "NumBitmaps", flist ) );
		bitmaps = (BITMAP16 **)malloc( head.NumBitmaps * sizeof( BITMAP16 * ) );
		bmphead = (BitmapHead *)malloc( head.NumBitmaps * sizeof( BitmapHead ) );
		if( !bitmaps || !bmphead ) {
			fclose( flist );
			bmp_free( tilebmp );
			bmp_free( fontbmp );
			free( bitmaps );
			free( bmphead );
			printf( "ERROR: Not enough memory for bitmap list!\n" );
			return 1;
		}
		for( i = 0; i < head.NumBitmaps; i++ ) {
			sprintf(	namebuf, "Bitmap%d", i );
			readline( linebuf, namebuf, flist );
			strtrim( linebuf );
			/* Get the bitmap name, then remove it from the buf */
			strncpy( bmphead[i].Name, strchr( linebuf, ';' )+1, 8 );
			*strchr( linebuf, ';' ) = 0;
			strtrim( linebuf );
			/* Load the bitmap */
			sprintf( filename, "%s%s", basepath, linebuf );
			bitmaps[i] = bmp_load( filename );
			if( !bitmaps[i] ) {
				fclose( flist );
				bmp_free( tilebmp );
				bmp_free( fontbmp );
				for( j = 0; j < i; j++ )
					bmp_free( bitmaps[j] );
				free( bitmaps );
				free( bmphead );
				printf( "ERROR: Cannot load %s!\n", filename );
				return 1;
			}
			bmphead[i].Width = bitmaps[i]->w / 8;
			bmphead[i].Height = bitmaps[i]->h;
		}

		/* Read the sprites */
		head.NumSprites = atoi( readline( linebuf, "NumSprites", flist ) );
		sprites = (BITMAP16 **)malloc( head.NumSprites * sizeof( BITMAP16 * ) );
		sprhead = (SpriteHead *)malloc( head.NumSprites * 4 * sizeof( SpriteHead ) );
		if( !sprites || !sprhead ) {
			fclose( flist );
			bmp_free( tilebmp );
			bmp_free( fontbmp );
			free( bitmaps );
			free( bmphead );
			free( sprites );
			free( sprhead );
			printf( "ERROR: Not enough memory for sprite list!\n" );
			return 1;
		}
		for( i = 0; i < head.NumSprites; i++ ) {
			sprintf(	namebuf, "Sprite%d", i );
			readline( linebuf, namebuf, flist );
			strtrim( linebuf );
			/* Get the sprite name, then remove it from the buf */
			strncpy( sprhead[i*4].Name, strchr( linebuf, ';' )+1, 16 );
			*strchr( linebuf, ';' ) = 0;
			strtrim( linebuf );
			/* Load the sprite */
			sprintf( filename, "%s%s", basepath, linebuf );
			sprites[i] = bmp_load( filename );
			if( !sprites[i] ) {
				fclose( flist );
				bmp_free( tilebmp );
				bmp_free( fontbmp );
				for( j = 0; j < head.NumBitmaps; j++ )
					bmp_free( bitmaps[j] );
				for( j = 0; j < i; j++ )
					bmp_free( sprites[j] );
				free( bitmaps );
				free( bmphead );
				free( sprites );
				free( sprhead );
				printf( "ERROR: Cannot load %s!\n", filename );
				return 1;
			}
			sprhead[i*4].Width = sprites[i]->w / 24;
			sprhead[i*4].Height = sprites[i]->h;
		}

		/* Fill out the header structure */
   	head.OffBitmapTable = sizeof( EgaHead );
   	head.OffSpriteTable = head.OffBitmapTable + head.NumBitmaps * sizeof( BitmapHead );
		head.Num32Tiles = 0;
		head.Off32Tiles = 0;
		head.Compressed = 0;

		/* Now calculate how much room all the data will take, working out
		offsets at the same time */
		head.LatchPlaneSize = 0;
		head.Off8Tiles = 0;
		head.LatchPlaneSize += (fontbmp->w * fontbmp->h) / 8;
		head.Off16Tiles = head.LatchPlaneSize;
		head.LatchPlaneSize += (tilebmp->w * tilebmp->h) / 8;
		head.OffBitmaps = head.LatchPlaneSize;
		for( i = 0; i < head.NumBitmaps; i++ ) {
			bmphead[i].Offset = head.LatchPlaneSize - head.OffBitmaps;
			head.LatchPlaneSize += (bitmaps[i]->w * bitmaps[i]->h) / 8;
		}
		head.SpritePlaneSize = 0;
		head.OffSprites = 0;
		for( i = 0; i < head.NumSprites; i++ ) {
			sprhead[i*4].OffsetParas = (head.SpritePlaneSize - head.OffSprites) / 16;
			sprhead[i*4].OffsetDelta = (head.SpritePlaneSize - head.OffSprites) % 16;
			head.SpritePlaneSize += (sprites[i]->w / 3 * sprites[i]->h) / 8;
		}

   	/* Backup then open EGAHEAD */
      sprintf( filename, "EGAHEAD.CK%c", episode + '0' );
      sprintf( bakfilename, "EGAHEAD.~K%c", episode + '0' );
		rename( filename, bakfilename );
      headfile = fopen( filename, "wb" );
      if( !headfile ) {
      	printf( "ERROR: Can't open %s! - is it readonly?\n", filename );
			fclose( flist );
			bmp_free( tilebmp );	bmp_free( fontbmp );
			for( j = 0; j < head.NumBitmaps; j++ )
				bmp_free( bitmaps[j] );
			for( j = 0; j < i; j++ )
				bmp_free( sprites[j] );
			free( bitmaps ); free( bmphead ); free( sprites );	free( sprhead );
         return 1;
      }

   	/* Backup then open EGALATCH */
      sprintf( filename, "EGALATCH.CK%c", episode + '0' );
      sprintf( bakfilename, "EGALATCH.~K%c", episode + '0' );
		rename( filename, bakfilename );
      latchfile = fopen( filename, "wb" );
      if( !latchfile ) {
      	printf( "ERROR: Can't open %s! - is it readonly?\n", filename );
			fclose( headfile );
			fclose( flist );
			bmp_free( tilebmp );	bmp_free( fontbmp );
			for( j = 0; j < head.NumBitmaps; j++ )
				bmp_free( bitmaps[j] );
			for( j = 0; j < i; j++ )
				bmp_free( sprites[j] );
			free( bitmaps ); free( bmphead ); free( sprites );	free( sprhead );
         return 1;
      }

   	/* Backup then open EGASPRIT */
      sprintf( filename, "EGASPRIT.CK%c", episode + '0' );
      sprintf( bakfilename, "EGASPRIT.~K%c", episode + '0' );
		rename( filename, bakfilename );
      spritefile = fopen( filename, "wb" );
      if( !spritefile ) {
      	printf( "ERROR: Can't open %s! - is it readonly?\n", filename );
			fclose( headfile ); fclose( latchfile );
			fclose( flist );
			bmp_free( tilebmp );	bmp_free( fontbmp );
			for( j = 0; j < head.NumBitmaps; j++ )
				bmp_free( bitmaps[j] );
			for( j = 0; j < i; j++ )
				bmp_free( sprites[j] );
			free( bitmaps ); free( bmphead ); free( sprites );	free( sprhead );
         return 1;
      }

		/* Now we call another function to decode all these bitmaps into planar
		data and write it to the files */
		import( &head, bmphead, sprhead, fontbmp, tilebmp, bitmaps, sprites,
 			headfile, latchfile, spritefile );


		/* Clean up! */
		fclose( headfile ); fclose( latchfile ); fclose( spritefile );
		fclose( flist ); bmp_free( tilebmp ); bmp_free( fontbmp );
		for( j = 0; j < head.NumBitmaps; j++ )
			bmp_free( bitmaps[j] );
		for( j = 0; j < i; j++ )
			bmp_free( sprites[j] );
		free( bitmaps ); free( bmphead ); free( sprites );	free( sprhead );

      /* And success */
      return 0;
   } else {
   	/* Invalid command */
     	printf( "ERROR: Unrecognised command '%c'\n", argv[1][0] );
      return 1;
   }


   return 0;
}

void import( EgaHead *head, BitmapHead *bmphead, SpriteHead *sprhead,
		BITMAP16 *fontbmp, BITMAP16 *tilebmp, BITMAP16 **bitmaps, BITMAP16 **sprites,
		FILE *headfile, FILE *latchfile, FILE *spritefile ) {
	int p, i, x, y, lx, ly;
	int latchpad, spritepad;

	/* Pad the planes to paragraphs */
	latchpad = (16 - (head->LatchPlaneSize % 16)) % 16;
	head->LatchPlaneSize += latchpad;

	/* Output the latch data */
	for( p = 0; p < 4; p++ ) {
		for( i = 0; i < head->Num8Tiles; i++ )
			decode_plane( 1 << p, latchfile, fontbmp, i%(fontbmp->w/8) * 8, i/(fontbmp->w/8) * 8, 8, 8 );
		for( i = 0; i < head->Num16Tiles; i++ )
			decode_plane( 1 << p, latchfile, tilebmp, i%(tilebmp->w/16) * 16, i/(tilebmp->w/16) * 16, 16, 16 );
		for( i = 0; i < head->NumBitmaps; i++ )
			decode_plane( 1 << p, latchfile, bitmaps[i], 0, 0, bitmaps[i]->w, bitmaps[i]->h );
		for( i = 0; i < latchpad; i++ )
			fputc( 0, latchfile );
	}

	/* Update the sprite header with clipping rectangle info etc */
	for( i = 0; i < head->NumSprites * 4; i+=4 ) {
		sprhead[i+3].Width = sprhead[i+2].Width = sprhead[i+1].Width = sprhead[i].Width + 1;
		sprhead[i+3].Height = sprhead[i+2].Height = sprhead[i+1].Height = sprhead[i].Height;
		sprhead[i+3].OffsetParas = sprhead[i+2].OffsetParas = sprhead[i+1].OffsetParas =	0;
		sprhead[i+3].OffsetDelta = sprhead[i+2].OffsetDelta = sprhead[i+1].OffsetDelta =	0;
		strncpy( sprhead[i+3].Name, sprhead[i].Name, 16 );
		strncpy( sprhead[i+2].Name, sprhead[i].Name, 16 );
		strncpy( sprhead[i+1].Name, sprhead[i].Name, 16 );
		/* Work out top-left corner of the clipping rectangle */
		lx = ly = 0;
		for( y = 0; y < sprites[i/4]->h; y++ ) {
			for( x = 0; x < sprites[i/4]->w / 3; x++ ) {
				p = bmp_getpixel( sprites[i/4], sprites[i/4]->w / 3 * 2 + x, y );
				if( p == 12 ) {
					lx = x;
					ly = y;
					goto foundtl;
				}
			}
		}
foundtl:
		sprhead[i+3].Rx1 = sprhead[i+2].Rx1 = sprhead[i+1].Rx1 = sprhead[i].Rx1 = lx << 8;
		sprhead[i+3].Ry1 = sprhead[i+2].Ry1 = sprhead[i+1].Ry1 = sprhead[i].Ry1 = ly << 8;

		/* Work out bottom-right corner of the clipping rectangle */
		lx = sprites[i/4]->w - 1;
		ly = sprites[i/4]->h - 1;
		for( y = 0; y < sprites[i/4]->h; y++ ) {
			for( x = 0; x < sprites[i/4]->w/3; x++ ) {
				p = bmp_getpixel( sprites[i/4], sprites[i/4]->w/3 * 2 + x, y );
				if( p == 12 ) {
					lx = x;
					ly = y;
				}
			}
		}
		sprhead[i+3].Rx2 = sprhead[i+2].Rx2 = sprhead[i+1].Rx2 = sprhead[i].Rx2 = lx << 8;
		sprhead[i+3].Ry2 = sprhead[i+2].Ry2 = sprhead[i+1].Ry2 = sprhead[i].Ry2 = ly << 8;
	}

	/* Pad the planes to paragraphs */
	spritepad = (16 - (head->SpritePlaneSize % 16)) % 16;
	head->SpritePlaneSize += spritepad;

	/* Output the sprite data */
	for( p = 0; p < 4; p++ ) {
		for( i = 0; i < head->NumSprites; i++ )
			decode_plane( 1 << p, spritefile, sprites[i], 0, 0, sprites[i]->w / 3, sprites[i]->h );
		for( i = 0;	i < spritepad; i++ )
			fputc( 0, spritefile );
	}
	for( i = 0; i < head->NumSprites; i++ )
		decode_plane( 15, spritefile, sprites[i], sprites[i]->w / 3, 0, sprites[i]->w / 3, sprites[i]->h );
	for( i = 0;	i < spritepad; i++ )
		fputc( 0, spritefile );

	/* Output the header */
	fwrite( head, sizeof( EgaHead ), 1, headfile );
	fwrite( bmphead, head->NumBitmaps * sizeof( BitmapHead ), 1, headfile );
	fwrite( sprhead, head->NumSprites * 4 * sizeof( SpriteHead ), 1, headfile );
}

void export( EgaHead *head, BitmapHead *bmphead, SpriteHead *sprhead,
		char *latchdata, char *spritedata, char *basepath, char *listfilename, int episode ) {
	BITMAP16 *bmp;
   FILE *listfile;
   SpriteHead *sh;
   char bmpname[PATH_MAX];
   char fullname[PATH_MAX];
   int i, x, y, p;

   /* Open the listfile for writing */
   listfile = fopen( listfilename, "wa" );
   if( !listfile ) {
   	printf( "ERROR: Can't create list file %s!\n", listfilename );
      return;
   }

   /* Print the list file header */
   fprintf( listfile, ";MODKEEN 1.0 list file\n" );
   fprintf( listfile, "Episode=%d\n", episode );

   /* Export all the 8x8 tiles into one bitmap */
   printf( "Exporting font...\n" );
   sprintf( bmpname, "%cFON0000.bmp", episode + '0' );
   fprintf( listfile, "\n;Font (8x8 tiles)\n" );
   fprintf( listfile, "Font=%s\n", bmpname );
   bmp = bmp_create( 16 * 8, (head->Num8Tiles + 15) / 16 * 8 );
   for( i = 0; i < head->Num8Tiles; i++ )
   	for( p = 0; p < 4; p++ )
         encode_plane( 1 << p, latchdata + p * head->LatchPlaneSize +
         	head->Off8Tiles + i * 8, bmp, i%16 * 8, i/16 * 8, 8, 8 );
   strcpy( fullname, basepath );
   strcat( fullname, bmpname );
	bmp_save( bmp, fullname );
   bmp_free( bmp );

   /* Export all the 16x16 tiles into one bitmap */
   printf( "Exporting tiles...\n" );
   sprintf( bmpname, "%cTIL0000.bmp", episode + '0' );
   fprintf( listfile, "\n;Tiles (16x16)\n" );
   fprintf( listfile, "Tiles=%s\n", bmpname );
   bmp = bmp_create( 13 * 16, (head->Num16Tiles + 12) / 13 * 16 );
   for( i = 0; i < head->Num16Tiles; i++ )
   	for( p = 0; p < 4; p++ )
         encode_plane( 1 << p, latchdata + p * head->LatchPlaneSize +
         	head->Off16Tiles + i * 32, bmp, i%13 * 16, i/13 * 16, 16, 16 );
   strcpy( fullname, basepath );
   strcat( fullname, bmpname );
	bmp_save( bmp, fullname );
   bmp_free( bmp );

   /* Export all the bitmaps */
   printf( "Exporting bitmaps" );
	fprintf( listfile, "\n;Bitmaps (different sizes)\n" );
	fprintf( listfile, "NumBitmaps=%d\n", head->NumBitmaps );
   for( i = 0; i < head->NumBitmaps; i++ ) {
	   printf( "." );
	   sprintf( bmpname, "%cBMP%04d.bmp", episode + '0', i );
	   fprintf( listfile, "Bitmap%d=%s ;%.8s\n", i, bmpname, bmphead[i].Name );
	   bmp = bmp_create( bmphead[i].Width*8, bmphead[i].Height );
   	for( p = 0; p < 4; p++ )
         encode_plane( 1 << p, latchdata + p * head->LatchPlaneSize +
         	head->OffBitmaps + bmphead[i].Offset, bmp, 0, 0, bmphead[i].Width*8, bmphead[i].Height );
	   strcpy( fullname, basepath );
	   strcat( fullname, bmpname );
		bmp_save( bmp, fullname );
	   bmp_free( bmp );
   }

   /* Export all the sprites */
   printf( "\nExporting sprites" );
	fprintf( listfile, "\n;Sprites (different sizes)\n" );
	fprintf( listfile, "NumSprites=%d\n", head->NumSprites );
   for( i = 0; i < head->NumSprites; i++ ) {
	   printf( "." );
	   sprintf( bmpname, "%cSPR%04d.bmp", episode + '0', i );
      sh = &sprhead[i*4];
	   fprintf( listfile, "Sprite%d=%s ;%.16s\n", i, bmpname, sh->Name );
	   bmp = bmp_create( sh->Width*8 * 3, sh->Height );
   	for( p = 0; p < 4; p++ )
         encode_plane( 1 << p, spritedata + p * head->SpritePlaneSize +
         	head->OffSprites + sh->OffsetParas * 16 + sh->OffsetDelta, bmp,
         	0, 0, sh->Width*8, sh->Height );
      /* Encode the mask */
      encode_plane( 15, spritedata + 4 * head->SpritePlaneSize +
       	sh->OffsetParas * 16 + sh->OffsetDelta, bmp,
        	sh->Width*8, 0, sh->Width*8, sh->Height );
      /* Draw the clipping rectangle */
      bmp_rect( bmp, 2*sh->Width*8, 0, 3*sh->Width*8 - 1,
      	sh->Height - 1, 8 );
      bmp_rect( bmp, 2*sh->Width*8 + (sh->Rx1>>8), (sh->Ry1>>8),
      	2*sh->Width*8 + (sh->Rx2>>8), (sh->Ry2>>8), 12 );
      /* Save the bitmap */
	   strcpy( fullname, basepath );
	   strcat( fullname, bmpname );
		bmp_save( bmp, fullname );
	   bmp_free( bmp );
   }
   printf( "\nDone!\n" );
}

char *readline( char *linebuf, char *tag, FILE *f ) {
	char *peq;

	do {
		fgets( linebuf, 256, f );
	} while( !feof( f ) && strnicmp( linebuf, tag, strlen( tag ) ) != 0 );

	peq = strchr( linebuf, '=' );
	if( peq )
		memmove( linebuf, peq+1, strlen( linebuf ) - (peq - linebuf) );

	return linebuf;
}

char *strtrim( char *buf ) {
	char *p = buf;
	while( isspace( *p ) )
		p++;
	memmove( buf, p, strlen( buf ) - (p - buf) + 1 );
	p = buf + strlen( buf ) - 1;
	while( p > buf && isspace( *p ) )
		p--;

	*(p+1) = 0;
}


void encode_plane( int cmask, char *pbits, BITMAP16 *bmp, int bx, int by, int w, int h ) {
	int tx, ty, x, y, c, p, pixel;
	int pmask;
	int b;

	/* Decode the data */
	for( y = 0; y < h; y++ ) {
		for( x = 0; x < w/8; x++ ) {
			b = *pbits++;

			/* Decode the byte */
			pmask = 128;
		   for( p = 0; p < 8; p++ ) {
				if( b & pmask ) {
					/* Shove a bit in the correct plane */
					pixel = bmp_getpixel( bmp, bx+x*8+p, by+y );
					pixel |= cmask;
					bmp_putpixel( bmp, bx+x*8+p, by+y, pixel );
				}
            pmask >>= 1;
			}
      }
	}
}


void decode_plane( int cmask, FILE *fout, BITMAP16 *bmp, int bx, int by,
	int w, int h ) {
	int c, x, y, b, p, pmask;

	/* There are four bitplanes */
	for( y = 0; y < h; y++ ) {
		for( x = 0; x < w/8; x++ ) {
			b = 0;

			/* Decode the byte */
			pmask = 128;
			for( p = 0; p < 8; p++ ) {
				if( bmp_getpixel( bmp, bx+x*8+p, by+y ) & cmask ) {
				  	/* Shove a bit in the correct plane */
					b |= pmask;
				}
         	pmask >>= 1;
        	}

			fputc( b, fout );
		}
	}
}



